import pytest
from cactus_test_definitions import variable_expressions as varexps
from cactus_test_definitions.client.events import Event, validate_event_parameters
from cactus_test_definitions.errors import TestProcedureDefinitionError


@pytest.mark.parametrize(
    "event, is_valid",
    [
        (Event("foo", {}), False),  # Unknown check
        (Event("wait", {}), False),
        (Event("wait", {"duration_seconds": "3"}), False),
        (Event("wait", {"duration_seconds": 3}), True),
        (Event("wait", {"duration_seconds": 3, "other": 4}), False),
        (Event("GET-request-received", {"endpoint": "/dcap"}), True),
        (Event("proceed", {}), True),
    ],
)
def test_validate_event_parameters(event: Event, is_valid: bool):
    if is_valid:
        validate_event_parameters("foo", "bar", event)
    else:
        with pytest.raises(TestProcedureDefinitionError):
            validate_event_parameters("foo", "bar", event)


def test_event_expression() -> None:
    """Tests the creation of an Event that has an expression as one of its parameters"""
    type_str = "some_event"
    params = {"setMaxW": "$(this < rtgMaxW)"}
    event = Event(type_str, params)

    check_set_max_w = event.parameters["setMaxW"]
    assert isinstance(check_set_max_w, varexps.Expression)
    assert check_set_max_w.operation == varexps.OperationType.LT
    assert check_set_max_w.lhs_operand == varexps.NamedVariable(varexps.NamedVariableType.DERSETTING_SET_MAX_W)
    assert check_set_max_w.rhs_operand == varexps.NamedVariable(varexps.NamedVariableType.DERCAPABILITY_RTG_MAX_W)



from dataclasses import dataclass
from enum import StrEnum
from importlib import resources

import yaml
from cactus_test_definitions.client.actions import Action
from cactus_test_definitions.client.checks import Check
from cactus_test_definitions.client.events import Event
from cactus_test_definitions.csipaus import CSIPAusVersion
from cactus_test_definitions.schema import UniqueKeyLoader
from dataclass_wizard import LoadMeta, YAMLWizard


class TestProcedureId(StrEnum):
    """The set of all available test ID's

    This should be kept in sync with the current set of client test procedures loaded from the procedures directory"""

    __test__ = False  # Prevent pytest from picking up this class
    ALL_01 = "ALL-01"
    ALL_02 = "ALL-02"
    ALL_03 = "ALL-03"
    ALL_03_REJ = "ALL-03-REJ"
    ALL_04 = "ALL-04"
    ALL_05 = "ALL-05"
    ALL_06 = "ALL-06"
    ALL_07 = "ALL-07"
    ALL_08 = "ALL-08"
    ALL_09 = "ALL-09"
    ALL_10 = "ALL-10"
    ALL_11 = "ALL-11"
    ALL_12 = "ALL-12"
    ALL_13 = "ALL-13"
    ALL_14 = "ALL-14"
    ALL_15 = "ALL-15"
    ALL_16 = "ALL-16"
    ALL_17 = "ALL-17"
    ALL_18 = "ALL-18"
    ALL_19 = "ALL-19"
    ALL_20 = "ALL-20"
    ALL_21 = "ALL-21"
    ALL_22 = "ALL-22"
    ALL_23 = "ALL-23"
    ALL_24 = "ALL-24"
    ALL_25 = "ALL-25"
    ALL_25_EXT = "ALL-25-EXT"
    ALL_26 = "ALL-26"
    ALL_27 = "ALL-27"
    ALL_28 = "ALL-28"
    ALL_29 = "ALL-29"
    ALL_30 = "ALL-30"
    DRA_01 = "DRA-01"
    DRA_02 = "DRA-02"
    DRD_01 = "DRD-01"
    DRL_01 = "DRL-01"
    DRG_01 = "DRG-01"
    GEN_01 = "GEN-01"
    GEN_02 = "GEN-02"
    GEN_03 = "GEN-03"
    GEN_04 = "GEN-04"
    GEN_05 = "GEN-05"
    GEN_06 = "GEN-06"
    GEN_07 = "GEN-07"
    GEN_08 = "GEN-08"
    GEN_09 = "GEN-09"
    GEN_10 = "GEN-10"
    GEN_11 = "GEN-11"
    GEN_12 = "GEN-12"
    GEN_13 = "GEN-13"
    LOA_01 = "LOA-01"
    LOA_02 = "LOA-02"
    LOA_03 = "LOA-03"
    LOA_04 = "LOA-04"
    LOA_05 = "LOA-05"
    LOA_06 = "LOA-06"
    LOA_07 = "LOA-07"
    LOA_08 = "LOA-08"
    LOA_09 = "LOA-09"
    LOA_10 = "LOA-10"
    LOA_11 = "LOA-11"
    LOA_12 = "LOA-12"
    LOA_13 = "LOA-13"
    MUL_01 = "MUL-01"
    MUL_02 = "MUL-02"
    MUL_03 = "MUL-03"

    # Storage extension
    STO_01 = "STO-01"
    STO_02 = "STO-02"
    STO_03 = "STO-03"
    STO_04 = "STO-04"
    STO_05 = "STO-05"
    STO_06 = "STO-06"


@dataclass
class Step:
    """A step is a part of the test procedure that waits for some form of event before running a set of actions.

    It's common for a step to activate other "steps" so that the state of the active test procedure can "evolve" in
    response to client behaviour

    Instructions are out-of-band operations that need performing during the step
    e.g. disconnect DER from grid, disable power consumption etc.
    """

    event: Event  # The event to act as a trigger
    actions: list[Action]  # The actions to execute when the trigger is met
    instructions: list[str] | None = None


@dataclass
class Preconditions:
    """Preconditions are run during the "initialization" state that precedes the start of a test. They typically
    allow for the setup of the test.

    Checks are also included to prevent a client from starting a test before they have correctly met preconditions

    Instructions are out-of-band operations that need performing at the start of the test procedure
    e.g. attach a load etc.

    If immediate_start is set to True - the "initialization" step will be progressed through immediately so that the
    client has no opportunity to interact with the server in this state. Any actions will still be executed. Do NOT
    utilise immediate_start with precondition checks.
    """

    init_actions: list[Action] | None = None  # To be executed as the runner starts (before anything can occur)
    immediate_start: bool = False  # If True - a test execution will have NO "pre-start" phase.
    actions: list[Action] | None = None  # To be executed as the test case "starts" (usually on request of client)
    checks: list[Check] | None = None  # Will prevent move from "init" state to "started" state of a test if any fail
    instructions: list[str] | None = None


@dataclass
class Criteria:
    """Criteria represent the final pass/fail analysis run after a TestProcedure completion. They can consider both
    the final state of the test system as well as the interactions that happened while it was running"""

    checks: list[Check] | None = None  # These should be run at test procedure finalization to determine pass/fail


@dataclass
class TestProcedure(YAMLWizard):
    """Top level object for collecting everything relevant to a single TestProcedure"""

    __test__ = False  # Prevent pytest from picking up this class
    description: str  # Metadata from test definitions
    category: str  # Metadata from test definitions
    classes: list[str]  # Metadata from test definitions
    target_versions: list[CSIPAusVersion]  # What version(s) of csip-aus is this test targeting?
    steps: dict[str, Step]
    preconditions: Preconditions | None = None  # These execute during "init" and setup the test for a valid start state
    criteria: Criteria | None = None  # How will success/failure of this procedure be determined?


LoadMeta(raise_on_unknown_json_key=True).bind_to(TestProcedure)


def parse_test_procedure(yaml_contents: str) -> TestProcedure:
    """Given a YAML string - parse a TestProcedure.

    This will ensure the YAML parser will use all the "strict" extensions to reduce the incidence of errors"""

    return TestProcedure.from_yaml(
        yaml_contents,
        decoder=yaml.load,  # type: ignore
        Loader=UniqueKeyLoader,
    )


def get_yaml_contents(test_procedure_id: TestProcedureId) -> str:
    """Finds the YAML contents for the TestProcedure with the specified TestProcedureId"""
    yaml_resource = resources.files("cactus_test_definitions.client.procedures") / f"{test_procedure_id}.yaml"
    with resources.as_file(yaml_resource) as yaml_file:
        with open(yaml_file, "r") as f:
            yaml_contents = f.read()
            return yaml_contents


def get_test_procedure(test_procedure_id: TestProcedureId) -> TestProcedure:
    """Gets the TestProcedure with the nominated ID by loading its definition from disk"""
    yaml_contents = get_yaml_contents(test_procedure_id)
    return parse_test_procedure(yaml_contents)


def get_all_test_procedures() -> dict[TestProcedureId, TestProcedure]:
    """Gets every TestProcedure, keyed by their TestProcedureId"""
    return {tp_id: get_test_procedure(tp_id) for tp_id in TestProcedureId}



from dataclasses import dataclass
from typing import Any

from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.parameters import (
    ParameterSchema,
    ParameterType,
    validate_parameters,
)
from cactus_test_definitions.variable_expressions import (
    parse_variable_expression_body,
    try_extract_variable_expression,
)


@dataclass
class Check:
    """A check represents some validation logic that runs during a Test Step and provides a pass/fail result with a
    description. It will typically inspect the state of the client based on what it has seen from the server

    eg: Ensuring that the client was able to see an EndDevice registration"""

    type: str
    parameters: dict[str, Any] = None  # type: ignore # This will be forced in __post_init__

    def __post_init__(self):
        """Some parameter values might contain variable expressions (eg: a string "$now") that needs to be replaced
        with an parsed Expression object instead."""
        if self.parameters is None:
            self.parameters = {}
        for k, v in self.parameters.items():
            variable_expr = try_extract_variable_expression(v)
            if variable_expr:
                self.parameters[k] = parse_variable_expression_body(variable_expr, k)


# The parameter schema for each action, keyed by the action name
CHECK_PARAMETER_SCHEMA: dict[str, dict[str, ParameterSchema]] = {
    "discovered": {
        "resources": ParameterSchema(False, ParameterType.ListCSIPAusResource),
        "links": ParameterSchema(False, ParameterType.ListCSIPAusResource),
    },
    "time-synced": {},  # Passes if the current Time resource is synced with this client's date/time
    "function-set-assignment": {
        "minimum_count": ParameterSchema(False, ParameterType.Integer),  # Needs at least this many FSAs to pass
        "maximum_count": ParameterSchema(False, ParameterType.Integer),  # Needs at most this many FSAs to pass
        "matches_client_edev": ParameterSchema(
            False, ParameterType.Boolean
        ),  # If True - only FSAs assigned to the client's EndDevice will be counted
        "sub_id": ParameterSchema(
            False, ParameterType.String
        ),  # If set - only FSAs received via this subscription will be counted
    },
    "end-device-list": {
        "minimum_count": ParameterSchema(False, ParameterType.Integer),  # Needs at least this many edev lists to pass
        "maximum_count": ParameterSchema(False, ParameterType.Integer),  # Needs at most this many edev lists to pass
        "poll_rate": ParameterSchema(
            False, ParameterType.Integer
        ),  # If set - will only count an EndDeviceList with this exact pollRate
        "sub_id": ParameterSchema(
            False, ParameterType.String
        ),  # If set - only EndDeviceLists received via this subscription will be counted
    },
    "end-device": {
        "matches_client": ParameterSchema(
            True, ParameterType.Boolean
        ),  # assert the existence / non existence of an EndDevice for the current client
        "matches_pin": ParameterSchema(
            False, ParameterType.Boolean
        ),  # if set - The matches_client criteria will ALSO check the registration PIN for the EndDevice. Default False
    },
    "der-program": {
        "minimum_count": ParameterSchema(False, ParameterType.Integer),  # Needs at least this many derps to pass
        "maximum_count": ParameterSchema(False, ParameterType.Integer),  # Needs at most this many derps to pass
        "primacy": ParameterSchema(False, ParameterType.Integer),  # Filters derps based on this primacy value
        "fsa_index": ParameterSchema(
            False, ParameterType.Integer
        ),  # Filters derps that belong to the nth (0 based) FunctionSetAssignment index
        "sub_id": ParameterSchema(
            False, ParameterType.String
        ),  # Filters derps to only those received via this named subscription
    },
    "der-control": {
        "minimum_count": ParameterSchema(False, ParameterType.Integer),  # Needs at least this many controls to pass
        "maximum_count": ParameterSchema(False, ParameterType.Integer),  # Needs at most this many controls to pass
        "latest": ParameterSchema(False, ParameterType.Boolean),  # forces filter checks against the most recent control
        "opModImpLimW": ParameterSchema(False, ParameterType.Float),  # Filters controls based on this value
        "opModExpLimW": ParameterSchema(False, ParameterType.Float),  # Filters controls based on this value
        "opModLoadLimW": ParameterSchema(False, ParameterType.Float),  # Filters controls based on this value
        "opModGenLimW": ParameterSchema(False, ParameterType.Float),  # Filters controls based on this value
        "opModEnergize": ParameterSchema(False, ParameterType.Boolean),  # Filters controls based on this value
        "opModConnect": ParameterSchema(False, ParameterType.Boolean),  # Filters controls based on this value
        "opModFixedW": ParameterSchema(False, ParameterType.Float),  # Filters controls based on this value
        "rampTms": ParameterSchema(False, ParameterType.Integer),  # Filter on this val. 0 means negative assertion
        "randomizeStart": ParameterSchema(False, ParameterType.Integer),  # Filter on this val (in seconds)
        "event_status": ParameterSchema(False, ParameterType.Integer),  # Filter on Event.status value
        "responseRequired": ParameterSchema(False, ParameterType.Integer),  # Filter on responseRequired value
        "derp_primacy": ParameterSchema(
            False, ParameterType.Integer
        ),  # Filter to control's belonging to a DERProgram with this primacy value
        "sub_id": ParameterSchema(
            False, ParameterType.String
        ),  # Filters control to only those received via this named subscription
        "duration": ParameterSchema(False, ParameterType.Integer),  # Filter on duration value
    },  # Matches many DERControls (specified by minimum_count) against additional other filter criteria
    "default-der-control": {
        "minimum_count": ParameterSchema(False, ParameterType.Integer),  # Needs at least this many default der controls
        "maximum_count": ParameterSchema(False, ParameterType.Integer),  # Needs at most this many default der controls
        "opModImpLimW": ParameterSchema(False, ParameterType.Float),
        "opModExpLimW": ParameterSchema(False, ParameterType.Float),
        "opModGenLimW": ParameterSchema(False, ParameterType.Float),
        "opModLoadLimW": ParameterSchema(False, ParameterType.Float),
        "setGradW": ParameterSchema(False, ParameterType.Integer),  # Hundredths of a percent / second
        "sub_id": ParameterSchema(
            False, ParameterType.String
        ),  # Filters default control to only those received via this named subscription
        "derp_primacy": ParameterSchema(
            False, ParameterType.Integer
        ),  # Filter to control's belonging to a DERProgram with this primacy value
    },  # matches any DefaultDERControl with the specified values
    "mirror-usage-point": {
        "matches": ParameterSchema(True, ParameterType.Boolean),  # True for positive assert, False for negative assert
        "check_mup_mrid": ParameterSchema(False, ParameterType.String),
        "location": ParameterSchema(False, ParameterType.CSIPAusReadingLocation),  # If not specified - match anything
        "reading_types": ParameterSchema(False, ParameterType.ListCSIPAusReadingType),  # If not specified - match all
        "mmr_mrids": ParameterSchema(
            False, ParameterType.ListString
        ),  # Must correspond 1-1 with reading_types. Used for forcing specific mrid values
        "post_rate_seconds": ParameterSchema(False, ParameterType.Integer),  # Only asserted if specified
    },  # True if the matches assertion finds a MirrorUsagePoint with the specified parameters (requires exact match)
    "subscription": {
        "matches": ParameterSchema(True, ParameterType.Boolean),  # True for positive assert, False for negative assert
        "resource": ParameterSchema(True, ParameterType.CSIPAusResource),
    },  # Matches the existence/nonexistence of a subscription for the specified resource
    "poll-rate": {
        "resource": ParameterSchema(True, ParameterType.CSIPAusResource),
        "poll_rate_seconds": ParameterSchema(True, ParameterType.Integer),
    },  # Asserts a specific poll rate value
}
VALID_CHECK_NAMES: set[str] = set(CHECK_PARAMETER_SCHEMA.keys())


def validate_check_parameters(procedure_name: str, check: Check) -> None:
    """Validates the check parameters for the parent TestProcedure based on the CHECK_PARAMETER_SCHEMA

    raises TestProcedureDefinitionError on failure"""
    location = f"{procedure_name} Check: {check.type}"  # Descriptive location of this action being validated

    parameter_schema = CHECK_PARAMETER_SCHEMA.get(check.type, None)
    if parameter_schema is None:
        raise TestProcedureDefinitionError(f"{location} not a valid action name. {VALID_CHECK_NAMES}")

    validate_parameters(location, check.parameters, parameter_schema)



class TestProcedureDefinitionError(Exception):
    __test__ = False  # Prevent pytest from picking up this class


class UnresolvableVariableError(Exception):
    """Raised whenever a NamedVariable cannot be resolved at test execution time (eg: database doesn't have the
    requisite information)"""

    pass


class UnparseableVariableExpressionError(Exception):
    """Raised whenever a raw string cannot parse to a NamedVariable/Expression"""

    pass



from pathlib import Path

import pytest
from cactus_test_definitions.client.test_procedures import (
    TestProcedureId,
    get_test_procedure,
    parse_test_procedure,
)
from cactus_test_definitions.client.validate import validate_test_procedure
from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.variable_expressions import (
    NamedVariableType,
    has_named_variable,
)


@pytest.mark.parametrize(
    "tp_file",
    [
        Path("tests/data/client/tp_invalid_bad_param.yaml"),
        Path("tests/data/client/tp_invalid_bad_step_enable.yaml"),
        Path("tests/data/client/tp_invalid_bad_step_remove.yaml"),
    ],
)
def test_TestProcedure_invalid_examples(tp_file: Path):
    with open(tp_file, "r") as fp:
        tp = parse_test_procedure(fp.read())

    with pytest.raises(TestProcedureDefinitionError):
        validate_test_procedure(tp, TestProcedureId.ALL_01)


@pytest.mark.parametrize("tp_id", TestProcedureId)
def test_TestProcedure_individually_valid(tp_id: TestProcedureId):
    """Tests that each TestProcedureId can be loaded via get_test_procedure without issue AND that it
    passes validate_test_procedure."""
    tp = get_test_procedure(tp_id)
    assert tp.steps, "There should be steps configured"
    validate_test_procedure(tp, tp_id)


@pytest.mark.parametrize("tp_id", TestProcedureId)
def test_each_step_accessible(tp_id: TestProcedureId):
    """Ensures that each test procedure's steps are enabled/removed once. Failures here indicate a bad reference to
    enable/remove steps"""

    tp = get_test_procedure(tp_id)

    first_step = [k for k in tp.steps.keys()][0]
    step_names = set(tp.steps.keys())

    removal_count = dict([(s, 0) for s in step_names])
    enabled_count = dict([(s, 0) for s in step_names])

    # Tally up the number of times each step is added/removed via actions
    for s in tp.steps.values():
        for enabled_name in [name for a in s.actions if a.type == "enable-steps" for name in a.parameters["steps"]]:
            enabled_count[enabled_name] = enabled_count[enabled_name] + 1
        for removed_name in [name for a in s.actions if a.type == "remove-steps" for name in a.parameters["steps"]]:
            removal_count[removed_name] = removal_count[removed_name] + 1

    # Also check the preconditions for actions
    if tp.preconditions and tp.preconditions.actions:
        for enabled_name in [
            name for a in tp.preconditions.actions if a.type == "enable-steps" for name in a.parameters["steps"]
        ]:
            enabled_count[enabled_name] = enabled_count[enabled_name] + 1
        for removed_name in [
            name for a in tp.preconditions.actions if a.type == "remove-steps" for name in a.parameters["steps"]
        ]:
            removal_count[removed_name] = removal_count[removed_name] + 1

    # Each test should be added once and removed once
    for step_name in step_names:
        assert (
            removal_count[step_name] == 1
        ), f"{step_name}: Each test should be removed once otherwise the test cannot complete"

        if step_name == first_step:
            assert enabled_count[step_name] == 0, f"{step_name}: The first test step is already enabled"
        else:
            assert (
                enabled_count[step_name] == 1
            ), f"{step_name}: Each test should be added once - not expecting loops of steps"


@pytest.mark.parametrize("tp_id", TestProcedureId)
def test_each_step_connected(tp_id: TestProcedureId):
    """Interprets each test a graph of steps linking to other steps (via enable-steps). Passes if the test procedure
    is fully connected from the first step (i.e. there are no isolated islands of steps that can't be reached from
    the first step)"""

    tp = get_test_procedure(tp_id)

    first_step = [k for k in tp.steps.keys()][0]
    step_names = set(tp.steps.keys())

    step_links_by_name: dict[str, list[str]] = {}
    for step_name in step_names:
        step = tp.steps[step_name]
        child_step_names = [name for a in step.actions if a.type == "enable-steps" for name in a.parameters["steps"]]
        step_links_by_name[step_name] = child_step_names

    # Walk through our simple graph/tree - make sure we visit every node in that tree
    def gather_connected_steps(links: dict[str, list[str]], node: str, visited_nodes: set[str]) -> None:
        if node in visited_nodes:
            return
        visited_nodes.add(node)
        for child in links[node]:
            gather_connected_steps(links, child, visited_nodes)

    visited_nodes: set[str] = set()
    gather_connected_steps(step_links_by_name, first_step, visited_nodes)

    # Also note that steps might be started as a precondition - so include those as visited too
    if tp.preconditions and tp.preconditions.actions:
        for n in [name for a in tp.preconditions.actions if a.type == "enable-steps" for name in a.parameters["steps"]]:
            visited_nodes.add(n)

    assert (
        step_names == visited_nodes
    ), "Missing entries here indicate a step (or steps) that can't be reached from the root node"


@pytest.mark.parametrize("tp_id", TestProcedureId)
def test_procedures_have_required_preconditions(tp_id: TestProcedureId):

    tp = get_test_procedure(tp_id)

    # Immediate start implies that EndDevice registration will happen during the test body - these tests
    # don't require end-device-contents (as it will cause the test to fail anyway)
    is_immediate_start = tp.preconditions is not None and tp.preconditions.immediate_start
    if not is_immediate_start:
        assert tp.preconditions is not None, "Expected precondition check 'end-device-contents'"
        assert tp.preconditions.checks is not None, "Expected precondition check 'end-device-contents'"
        assert any(
            [check.type == "end-device-contents" for check in tp.preconditions.checks]
        ), "Expected precondition check 'end-device-contents'"

    # Check 'der-settings-contents' present if any precondition action parameter references setMaxW
    if tp.preconditions is not None and tp.preconditions.actions is not None:
        for action in tp.preconditions.actions:
            if action.parameters is not None:
                if any(
                    [
                        has_named_variable(
                            parameter_value=parameter, named_variable=NamedVariableType.DERSETTING_SET_MAX_W
                        )
                        for parameter in action.parameters.values()
                    ]
                ):
                    assert tp.preconditions.checks is not None
                    assert any([check.type == "der-settings-contents" for check in tp.preconditions.checks])


@pytest.mark.parametrize("tp_id", TestProcedureId)
def test_response_subject_tags_have_corresponding_der_control_tags(tp_id: TestProcedureId):
    """Ensures that every subject_tag in response-contents checks has a corresponding
    tag defined in a create-der-control action in the same test procedure."""

    tp = get_test_procedure(tp_id)

    # Collect all tags defined in create-der-control actions
    der_control_tags = set()

    if tp.preconditions and tp.preconditions.actions:
        der_control_tags.update(
            [
                a.parameters.get("tag")
                for a in tp.preconditions.actions
                if a.type == "create-der-control" and a.parameters and a.parameters.get("tag")
            ]
        )

    der_control_tags.update(
        [
            a.parameters.get("tag")
            for s in tp.steps.values()
            for a in s.actions
            if a.type == "create-der-control" and a.parameters and a.parameters.get("tag")
        ]
    )

    # Collect all subject_tags referenced in response-contents checks
    referenced_subject_tags = set(
        [
            c.parameters.get("subject_tag")
            for s in tp.steps.values()
            if s.event.checks
            for c in s.event.checks
            if c.type == "response-contents" and c.parameters and c.parameters.get("subject_tag")
        ]
    )

    # Verify all referenced subject_tags have corresponding der-control tags
    missing_tags = referenced_subject_tags - der_control_tags
    assert not missing_tags, (
        f"The following subject_tags in response-contents checks do not have corresponding "
        f"tags in create-der-control actions: {missing_tags}"
    )



from cactus_test_definitions.client.actions import ACTION_PARAMETER_SCHEMA, Action
from cactus_test_definitions.client.checks import CHECK_PARAMETER_SCHEMA, Check
from cactus_test_definitions.client.events import EVENT_PARAMETER_SCHEMA, Event
from cactus_test_definitions.client.test_procedures import (
    Preconditions,
    Step,
    TestProcedure,
    TestProcedureId,
    get_all_test_procedures,
    get_test_procedure,
    get_yaml_contents,
    parse_test_procedure,
)

__all__ = [
    "TestProcedureId",
    "Event",
    "Action",
    "ACTION_PARAMETER_SCHEMA",
    "Check",
    "CHECK_PARAMETER_SCHEMA",
    "EVENT_PARAMETER_SCHEMA",
    "Step",
    "Preconditions",
    "TestProcedure",
    "get_all_test_procedures",
    "get_test_procedure",
    "get_yaml_contents",
    "parse_test_procedure",
]



from dataclasses import dataclass
from typing import Any

from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.parameters import (
    ParameterSchema,
    ParameterType,
    validate_parameters,
)
from cactus_test_definitions.variable_expressions import (
    parse_variable_expression_body,
    try_extract_variable_expression,
)


@dataclass
class Check:
    """A check represents some validation logic that runs during test finalization and can provide a pass/fail
    status beyond the "basic" flow of a test procedure. Checks will typically inspect the database/history of requests
    ino order to determine compliance.

    eg: Looking through past requests to an endpoint to validate it was called at a specific rate"""

    type: str
    parameters: dict[str, Any]

    def __post_init__(self):
        """Some parameter values might contain variable expressions (eg: a string "$now") that needs to be replaced
        with an parsed Expression object instead."""
        for k, v in self.parameters.items():
            variable_expr = try_extract_variable_expression(v)
            if variable_expr:
                self.parameters[k] = parse_variable_expression_body(variable_expr, k)


def factory_readings_schema() -> dict[str, Any]:
    """Factory function for common schema shared across reading checks."""
    return {
        "minimum_count": ParameterSchema(False, ParameterType.Integer),
        "minimum_level": ParameterSchema(False, ParameterType.Float),
        "maximum_level": ParameterSchema(False, ParameterType.Float),
        "window_seconds": ParameterSchema(False, ParameterType.UnsignedInteger),
    }


# The parameter schema for each action, keyed by the action name
CHECK_PARAMETER_SCHEMA: dict[str, dict[str, ParameterSchema]] = {
    "all-steps-complete": {"ignored_steps": ParameterSchema(False, ParameterType.ListString)},
    "all-notifications-transmitted": {},
    "end-device-contents": {
        "has_connection_point_id": ParameterSchema(False, ParameterType.Boolean),
        "deviceCategory_anyset": ParameterSchema(False, ParameterType.HexBinary),  # Any of these bits set to 1
        "check_lfdi": ParameterSchema(False, ParameterType.Boolean),  # Should LFDI be validated in detail
    },
    "der-settings-contents": {
        "setGradW": ParameterSchema(False, ParameterType.Integer),  # Hundredths of a percent / second
        "doeModesEnabled": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "doeModesEnabled_set": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to one
        "doeModesEnabled_unset": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to zero
        "modesEnabled_set": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to one
        "modesEnabled_unset": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to zero
        "vppModesEnabled": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "vppModesEnabled_set": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to one
        "vppModesEnabled_unset": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to zero
        "setMaxVA": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "setMaxVar": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "setMaxVarNeg": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "setMaxW": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "setMaxChargeRateW": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "setMaxDischargeRateW": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "setMaxWh": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "setMinWh": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "setMinPFOverExcited": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "setMinPFUnderExcited": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
    },
    "der-capability-contents": {
        "doeModesSupported": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "doeModesSupported_set": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to one
        "doeModesSupported_unset": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to zero
        "modesSupported_set": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to one
        "modesSupported_unset": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to zero
        "vppModesSupported": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "vppModesSupported_set": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to one
        "vppModesSupported_unset": ParameterSchema(False, ParameterType.HexBinary),  # Minimum bits set to zero
        "rtgMaxVA": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "rtgMaxVar": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "rtgMaxVarNeg": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "rtgMaxW": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "rtgMaxChargeRateW": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "rtgMaxDischargeRateW": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "rtgMaxWh": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "rtgMinPFOverExcited": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
        "rtgMinPFUnderExcited": ParameterSchema(False, ParameterType.Boolean),  # Is ANY value set?
    },
    "der-status-contents": {
        "genConnectStatus": ParameterSchema(False, ParameterType.Integer),
        "genConnectStatus_bit0": ParameterSchema(False, ParameterType.Boolean),
        "genConnectStatus_bit1": ParameterSchema(False, ParameterType.Boolean),
        "genConnectStatus_bit2": ParameterSchema(False, ParameterType.Boolean),
        "operationalModeStatus": ParameterSchema(False, ParameterType.Integer),
        "alarmStatus": ParameterSchema(False, ParameterType.Integer),
    },
    "readings-voltage": factory_readings_schema(),
    "readings-site-active-power": factory_readings_schema(),
    "readings-site-reactive-power": factory_readings_schema(),
    "readings-der-active-power": factory_readings_schema(),
    "readings-der-reactive-power": factory_readings_schema(),
    "readings-der-stored-energy": factory_readings_schema(),
    "subscription-contents": {"subscribed_resource": ParameterSchema(True, ParameterType.String)},
    "response-contents": {
        "latest": ParameterSchema(False, ParameterType.Boolean),
        "status": ParameterSchema(False, ParameterType.Integer),
        "all": ParameterSchema(False, ParameterType.Boolean),
        "subject_tag": ParameterSchema(False, ParameterType.String),
    },
    "all-polls-at-correct-time": {
        "endpoint": ParameterSchema(True, ParameterType.String),  # e.g. /dcap
        "poll_interval_seconds": ParameterSchema(True, ParameterType.Integer),
        "request_type_str": ParameterSchema(True, ParameterType.String),  # e.g. GET, POST
    },
}
VALID_CHECK_NAMES: set[str] = set(CHECK_PARAMETER_SCHEMA.keys())


def validate_check_parameters(procedure_name: str, check: Check) -> None:
    """Validates the check parameters for the parent TestProcedure based on the CHECK_PARAMETER_SCHEMA

    raises TestProcedureDefinitionError on failure"""
    location = f"{procedure_name} Check: {check.type}"  # Descriptive location of this action being validated

    parameter_schema = CHECK_PARAMETER_SCHEMA.get(check.type, None)
    if parameter_schema is None:
        raise TestProcedureDefinitionError(f"{location} not a valid check name. {VALID_CHECK_NAMES}")

    validate_parameters(location, check.parameters, parameter_schema)
